/* -*- mode: c++; compile-command: "BOARD=pro5v make"; -*- */
/*
 * propeller-clock.ino
 *
 * Copyright (C) 2011 Tim Marston <tim@ed.am> and Dan Marston.
 *
 * This file is part of propeller-clock (hereafter referred to as "this
 * program"). See http://ed.am/dev/software/arduino/propeller-clock for more
 * information.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/******************************************************************************

Set up:

 * a PC fan is wired up to a 12V power supply

 * the fan's SENSE (tachiometer) pin connected to pin 2 on the
   arduino.

 * the pins 4 to 13 on the arduino should directly drive an LED (the
   LED on pin 4 is in the centre of the clock face and the LED on pin
   13 is at the outside.

 * if a longer hand (and a larger clock face) is desired, pin 4 can be
   used to indirectly drive a transistor which in turn drives several
   LEDs that turn on anf off in unison in the centre of the clock.

 * a button should be attached to pin 3 that grounds it when pressed.

 * A DS1307 remote clock is connected via I2C on analog pins 4 and 5.

Implementation details:

 * for a schematic, see ../project/propeller-clock.sch.

 * the timing of the drawing of the clock face is recalculated with
   every rotation of the propeller.
    
 * a PC fan actually sends 2 tachiometer pulses per revolution, so the
   software skips every other one. This means that the clock may
   appear upside-down if started with the propeller in the wrong
   position. You will need to experiment to dicsover the position that
   the propeller must be in when starting the clock.
    
Usage instructions:

 * pressing the button cycles between variations of the current
   display mode.
  
 * pressing and holding the button for a second cycles between display
   modes (e.g., analogue and digital).

 * pressing and holding the button for 5 seconds enters "time set"
   mode. In this mode, the following applies:
    - the field that is being set flashes
    - pressing the button increments the field currently being set
    - pressing and holding the button for a second cycles through the
      fields that can be set
    - pressing and holding the button for 5 seconds sets the time and
      exits "time set" mode

******************************************************************************/

#include "config.h"
#include "display.h"
#include "button.h"
#include "time.h"
#include "switcher_major_mode.h"
#include "drawer.h"

//_____________________________________________________________________________
//                                                                         data


// when non-zero, the time (in microseconds) of a new fan pulse that
// has just occurred, which means that segment drawing needs to be
// restarted
static unsigned long new_pulse_at = 0;

// the time (in microseconds) when the last fan pulse occurred
static unsigned long last_pulse_at = 0;

// duration (in microseconds) that a segment should be displayed
static unsigned long segment_step = 0;

// remainder after divisor and a tally of the remainders for each segment
static unsigned long segment_step_sub_step = 0;
static unsigned long segment_step_sub = 0;

// the button
static Button button( 3 );

// major mode
static int major_mode = 0;

#define MAX_MAJOR_MODES 5

// major modes
static MajorMode *major_modes[ MAX_MAJOR_MODES ] = { 0 };

//_____________________________________________________________________________
//                                                                         code


// perform button events
void doButtonEvents()
{
	// loop through pending events
	while( int event = button.get_event() )
	{
		switch( event )
		{
		case 1:
			// short press
			major_modes[ major_mode ]->press();
			break;

		case 2:
			// long press
			major_modes[ major_mode ]->long_press();
			break;

		case 3:
			// looooong press (change major mode)
			do {
				if( ++major_mode >= MAX_MAJOR_MODES )
					major_mode = 0;
			} while( major_modes[ major_mode ] == NULL );
			major_modes[ major_mode ]->activate();
			break;

		}
	}
}


// draw a display segment
void drawNextSegment( bool reset )
{
	// keep track of segment
#if CLOCK_FORWARD
	static int segment = ( NUM_SEGMENTS - CLOCK_SHIFT ) % NUM_SEGMENTS;
	if( reset ) segment = ( NUM_SEGMENTS - CLOCK_SHIFT ) % NUM_SEGMENTS;
#else
	static int segment = NUM_SEGMENTS - 1 - CLOCK_SHIFT;
	if( reset ) segment = NUM_SEGMENTS - 1 - CLOCK_SHIFT;
#endif

	// draw
	Drawer &drawer = major_modes[ major_mode ]->get_drawer();
	if( reset ) drawer.draw_reset();
	drawer.draw( segment );

#if CLOCK_FORWARD
	if( ++segment >= NUM_SEGMENTS ) segment = 0;
#else
	if( --segment < 0 ) segment = NUM_SEGMENTS - 1;
#endif
}


// calculate time constants when a new pulse has occurred
void calculateSegmentTimes()
{
	// check for overflows, and only recalculate times if there isn't
	// one (if there is, we'll just go with the last pulse's times)
	if( new_pulse_at > last_pulse_at )
	{
		// new segment stepping times
		unsigned long delta = new_pulse_at - last_pulse_at;
		segment_step = delta / NUM_SEGMENTS;
		segment_step_sub = 0;
		segment_step_sub_step = delta % NUM_SEGMENTS;
	}

	// now we have dealt with this pulse, save the pulse time and
	// clear new_pulse_at, ready for the next pulse
	last_pulse_at = new_pulse_at;
	new_pulse_at = 0;
}


// wait until it is time to draw the next segment or a new pulse has
// occurred
void waitTillEndOfSegment( bool reset )
{
	static unsigned long end_time = 0;

	// handle reset
	if( reset )
		end_time = last_pulse_at;

	// work out the time that this segment should be displayed until
	end_time += segment_step;
	segment_step_sub += segment_step_sub_step;
	if( segment_step_sub >= NUM_SEGMENTS ) {
		segment_step_sub -= NUM_SEGMENTS;
		end_time++;
	}

	// wait
	while( micros() < end_time && !new_pulse_at );
}


// ISR to handle the pulses from the fan's tachiometer
void fanPulseHandler()
{
	// the fan actually sends two pulses per revolution. These pulses
	// may not be exactly evenly distributed around the rotation, so
	// we can't recalculate times on every pulse. Instead, we ignore
	// every other pulse so timings are based on a complete rotation.
	static bool ignore = true;
	ignore = !ignore;
	if( !ignore )
	{
		// set a new pulse time
		new_pulse_at = micros();
	}
}


// main setup
void setup()
{
	// set up an interrupt handler on pin 2 to nitice fan pulses
	attachInterrupt( 0, fanPulseHandler, RISING );
	digitalWrite( 2, HIGH );
  
	// set up output pins (4 to 13) for the led array
	for( int a = 4; a < 14; a++ )
		pinMode( a, OUTPUT );

	// set up mode-switch button on pin 3
	pinMode( 3, INPUT );
	digitalWrite( 3, HIGH );
	static int event_times[] = { 5, 1000, 4000, 0 };
	button.set_event_times( event_times );

	// set up major modes
	static SwitcherMajorMode switcher_major_mode;
	int mode = 0;
	major_modes[ mode++ ] = &switcher_major_mode;
	major_modes[ 0 ]->activate();
}


// main loop
void loop()
{
	// if there has been a new pulse, we'll be resetting the display
	bool reset = new_pulse_at? true : false;

	// update button
	button.update();

	// only do this stuff at the start of a display cycle, to ensure
	// that no state changes mid-display
	if( reset )
	{
		// calculate segment times
		calculateSegmentTimes();

		// keep track of time
		Time &time = Time::get_instance();
		time.update();

		// perform button events
		doButtonEvents();
	}

	// draw this segment
	drawNextSegment( reset );

	// wait till it's time to draw the next segment
	waitTillEndOfSegment( reset );
}
